#include <linux/types.h>
#include <linux/bootinfo.h>
#include <linux/interrupt.h>
#include <linux/amigaints.h>
#include <linux/amigahw.h>
#include "../block/blk.h"
#include "scsi.h"
#include "hosts.h"
#include "wd33c93.h"
#include "a2091.h"

#define DMA(ptr) ((a2091_scsiregs *)((ptr)->base))

static struct Scsi_Host *first_instance = NULL;
static Scsi_Host_Template *a2091_template;

static void a2091_intr (struct intframe *fp, void *data)
{
    unsigned int status;
    struct Scsi_Host *instance;
    unsigned long flags;

    for (instance = first_instance; instance &&
	 instance->hostt == a2091_template; instance = instance->next)
    {
	status = DMA(instance)->ISTR;
	if (!(status & (ISTR_INT_F|ISTR_INT_P)))
	    continue;

	if (status & ISTR_INTS)
	{
	    /* disable PORTS interrupt */
	    custom.intena = IF_PORTS;
	    save_flags (flags);
	    WD33C93_intr (instance);
	    restore_flags (flags);
	    /* enable PORTS interrupt */
	    custom.intena = IF_SETCLR | IF_PORTS;
	}
    }
}

static int dma_setup (Scsi_Cmnd *cmd, int dir_in)
{
    unsigned short cntr = CNTR_PDMD | CNTR_INTEN;
    unsigned long addr = VTOP(cmd->SCp.ptr);

    /* setup dma direction */
    if (!dir_in)
	cntr |= CNTR_DDIR;
    ((struct WD33C93_hostdata *)(cmd->host->hostdata))->dma_dir = dir_in;
    DMA(cmd->host)->CNTR = cntr;

    /* setup DMA *physical* address */
    DMA(cmd->host)->ACR = addr;

    if (dir_in)
	/* invalidate any cache */
	cache_clear (addr, cmd->SCp.this_residual);
    else
	/* push any dirty cache */
	cache_push (addr, cmd->SCp.this_residual);

    /* start DMA */
    DMA(cmd->host)->ST_DMA = 1;

    /* return success */
    return 0;
}

static void dma_stop (Scsi_Cmnd *SCpnt)
{
    /* disable SCSI interrupts */
    DMA(SCpnt->host)->CNTR = CNTR_PDMD;

    /* flush if we were reading */
    if (((struct WD33C93_hostdata *)(SCpnt->host->hostdata))->dma_dir) {
	DMA(SCpnt->host)->FLUSH = 1;
	while (!(DMA(SCpnt->host)->ISTR & ISTR_FE_FLG))
	    ;
    }

    /* clear a possible interrupt */
    DMA(SCpnt->host)->CINT = 1;

    /* stop DMA */
    DMA(SCpnt->host)->SP_DMA = 1;

    /* restore the CONTROL bits (minus the direction flag) */
    DMA(SCpnt->host)->CNTR = CNTR_PDMD | CNTR_INTEN;
}

int a2091_detect(int host_nr)
{
    static unsigned char called = 0;
    struct Scsi_Host *instance;
    int i, manuf, product, num_a2091 = 0;
    caddr_t address;

    if (called)
	return 0;
    called = 1;

    for (i = 0; i < boot_info.bi_amiga.num_autocon; i++)
    {
	manuf = boot_info.bi_amiga.autocon[i].cd_Rom.er_Manufacturer;
	product = boot_info.bi_amiga.autocon[i].cd_Rom.er_Product;
	if (manuf == MANUF_COMMODORE && (product == PROD_A2091 ||
					 product == PROD_A590)) {
	    address = boot_info.bi_amiga.autocon[i].cd_BoardAddr;
	    instance = scsi_register (host_nr,
				      sizeof (struct WD33C93_hostdata));
	    instance->base = (unsigned char *)address;
	    DMA(instance)->DAWR = DAWR_A2091;
	    wd33c93_init(instance, (wd33c93_regs *)&(DMA(instance)->SASR),
			 dma_setup, dma_stop, A2091_XFER_MASK,
			 WD33C93_FS_8_10);
	    host_nr++;
	    if (num_a2091++ == 0) {
		first_instance = instance;
		a2091_template = instance->hostt;
		add_isr(IRQ_AMIGA_PORTS, a2091_intr, 0, NULL);
	    }
	    DMA(instance)->CNTR = CNTR_PDMD | CNTR_INTEN;
	}
    }

    return num_a2091;
}
