#include <linux/types.h>
#include <linux/bootinfo.h>
#include <linux/interrupt.h>
#include <linux/amigaints.h>
#include <linux/amigahw.h>
#include "../block/blk.h"
#include "scsi.h"
#include "hosts.h"
#include "wd33c93.h"
#include "a3000.h"

#define DMA(ptr) ((a3000_scsiregs *)((ptr)->base))

static struct Scsi_Host *a3000_host = NULL;

static void a3000_intr (struct intframe *fp, void *data)
{
    unsigned int status = DMA(a3000_host)->ISTR;
    unsigned long flags;

    if (!(status & (ISTR_INT_F|ISTR_INT_P)))
	return;

    if (status & ISTR_INTS)
    {
	/* disable PORTS interrupt */
	custom.intena = IF_PORTS;
	save_flags (flags);
	WD33C93_intr (a3000_host);
	restore_flags (flags);
	/* enable PORTS interrupt */
	custom.intena = IF_SETCLR | IF_PORTS;
    }
}

static int dma_setup (Scsi_Cmnd *cmd, int dir_in)
{
	unsigned short cntr = CNTR_PDMD | CNTR_INTEN;
	unsigned long addr = VTOP(cmd->SCp.ptr);

	/* setup dma direction */
	if (!dir_in)
		cntr |= CNTR_DDIR;
	((struct WD33C93_hostdata *)(a3000_host->hostdata))->dma_dir
		= dir_in;
	DMA(a3000_host)->CNTR = cntr;

	/* setup DMA *physical* address */
	DMA(a3000_host)->ACR = addr;

	if (dir_in)
		/* invalidate any cache */
		cache_clear (addr, cmd->SCp.this_residual);
	else
		/* push any dirty cache */
		cache_push (addr, cmd->SCp.this_residual);

	/* start DMA */
	DMA(a3000_host)->ST_DMA = 1;

	/* return success */
	return 0;
}

static void dma_stop (Scsi_Cmnd *SCpnt)
{
	/* disable SCSI interrupts */
	DMA(a3000_host)->CNTR = CNTR_PDMD;

	/* flush if we were reading */
	if (((struct WD33C93_hostdata *)(a3000_host->hostdata))->dma_dir) {
		DMA(a3000_host)->FLUSH = 1;
		while (!(DMA(a3000_host)->ISTR & ISTR_FE_FLG))
			;
	}

	/* clear a possible interrupt */
	DMA(a3000_host)->CINT = 1;

	/* stop DMA */
	DMA(a3000_host)->SP_DMA = 1;

	/* restore the CONTROL bits (minus the direction flag) */
	DMA(a3000_host)->CNTR = CNTR_PDMD | CNTR_INTEN;
}

int a3000_detect(int host_nr)
{
    static unsigned char called = 0;

    if (called)
	return 0;

    if (boot_info.bi_amiga.model != AMI_3000)
	return 0;

    a3000_host = scsi_register (host_nr, sizeof(struct WD33C93_hostdata));
    a3000_host->base = (unsigned char *)0xDD0000;
    DMA(a3000_host)->DAWR = DAWR_A3000;
    wd33c93_init(a3000_host, (wd33c93_regs *)&(DMA(a3000_host)->SASR),
		 dma_setup, dma_stop, A3000_XFER_MASK, WD33C93_FS_12_15);
    add_isr(IRQ_AMIGA_PORTS, a3000_intr, 0, NULL);
    DMA(a3000_host)->CNTR = CNTR_PDMD | CNTR_INTEN;
    called = 1;

    return 1;
}
