#include <linux/types.h>
#include <linux/bootinfo.h>
#include <linux/interrupt.h>
#include <linux/amigaints.h>
#include <linux/amigahw.h>
#include "../block/blk.h"
#include "scsi.h"
#include "hosts.h"
#include "wd33c93.h"
#include "gvp11.h"

#define GVP_DPRC_IOEXTENDER   (0xb)

#define DMA(ptr) ((gvp11_scsiregs *)((ptr)->base))

static struct Scsi_Host *first_instance = NULL;
static Scsi_Host_Template *gvp11_template;

static void gvp11_intr (struct intframe *fp, void *data)
{
    unsigned int status;
    struct Scsi_Host *instance;
    unsigned long flags;

    for (instance = first_instance; instance &&
	 instance->hostt == gvp11_template; instance = instance->next)
    {
	status = DMA(instance)->CNTR;
	if (!(status & GVP11_DMAC_INT_PENDING))
	    continue;

	/* disable PORTS interrupt */
	custom.intena = IF_PORTS;
	save_flags (flags);
	WD33C93_intr (instance);
	restore_flags (flags);
	/* enable PORTS interrupt */
	custom.intena = IF_SETCLR | IF_PORTS;
    }
}

static int dma_setup (Scsi_Cmnd *cmd, int dir_in)
{
    unsigned short cntr = GVP11_DMAC_INT_ENABLE;
    unsigned long addr = VTOP(cmd->SCp.ptr);

    /* setup dma direction */
    if (!dir_in)
	cntr |= GVP11_DMAC_DIR_WRITE;
    ((struct WD33C93_hostdata *)(cmd->host->hostdata))->dma_dir = dir_in;
    DMA(cmd->host)->CNTR = cntr;

    /* setup DMA *physical* address */
    DMA(cmd->host)->ACR = addr;

    if (dir_in)
	/* invalidate any cache */
	cache_clear (addr, cmd->SCp.this_residual);
    else
	/* push any dirty cache */
	cache_push (addr, cmd->SCp.this_residual);

    /* start DMA */
    DMA(cmd->host)->ST_DMA = 1;

    /* return success */
    return 0;
}

static void dma_stop (Scsi_Cmnd *SCpnt)
{
    /* stop DMA */
    DMA(SCpnt->host)->SP_DMA = 1;
    /* remove write bit from CONTROL bits */
    DMA(SCpnt->host)->CNTR = GVP11_DMAC_INT_ENABLE;
}

int gvp11_detect(int host_nr)
{
    static unsigned char called = 0;
    struct Scsi_Host *instance;
    int i, manuf, product, num_gvp11 = 0;
    caddr_t address;
    unsigned short epc;

    if (called)
	return 0;
    called = 1;

    for (i = 0; i < boot_info.bi_amiga.num_autocon; i++)
    {
	manuf = boot_info.bi_amiga.autocon[i].cd_Rom.er_Manufacturer;
	product = boot_info.bi_amiga.autocon[i].cd_Rom.er_Product;
	if (manuf == MANUF_GVP && product == PROD_GVPIISCSI) {
	    /* check extended product code */
	    address = boot_info.bi_amiga.autocon[i].cd_BoardAddr;
	    epc = *(unsigned short *)(address + 0x8000);

	    if ((epc & 0x0f) == GVP_DPRC_IOEXTENDER)
		continue;

	    instance = scsi_register (host_nr,
				      sizeof (struct WD33C93_hostdata));
	    instance->base = (unsigned char *)address;
	    DMA(instance)->secret2 = 1;
	    DMA(instance)->secret1 = 0;
	    DMA(instance)->secret3 = 15;
	    while (DMA(instance)->CNTR & GVP11_DMAC_BUSY) ;
	    DMA(instance)->CNTR = 0;
	    wd33c93_init(instance, (wd33c93_regs *)&(DMA(instance)->SASR),
			 dma_setup, dma_stop, GVP11_XFER_MASK,
			 WD33C93_FS_8_10);
	    host_nr++;
	    if (num_gvp11++ == 0) {
		first_instance = instance;
		gvp11_template = instance->hostt;
		add_isr(IRQ_AMIGA_PORTS, gvp11_intr, 0, NULL);
	    }
	    DMA(instance)->CNTR = GVP11_DMAC_INT_ENABLE;
	}
    }

    return num_gvp11;
}
