/*
** lnux/amigatypes.h -- Types used in Amiga Linux kernel source
**
** Copyright 1992 by Greg Harp
**
** This file is subject to the terms and conditions of the GNU General Public
** License.  See the file README.legal in the main directory of this archive
** for more details.
**
** Created 09/29/92 by Greg Harp
*/

#ifndef _LINUX_AMIGATYPES_H_
#define _LINUX_AMIGATYPES_H_

struct Node {
    struct  Node *ln_Succ;	/* Pointer to next (successor) */
    struct  Node *ln_Pred;	/* Pointer to previous (predecessor) */
    u_char   ln_Type;
    char    ln_Pri;		/* Priority, for sorting */
    char    *ln_Name;		/* ID string, null terminated */
};

/* defined board manufacturers */
#define MANUF_COMMODORE      (0x202)
#define MANUF_GVP            (2017)

/* defined board product numbers */
#define PROD_A2091           (3)
#define PROD_A590            (2)
#define PROD_GVPIISCSI       (11)

struct ExpansionRom {
    /* -First 16 bytes of the expansion ROM */
    u_char	er_Type;	/* Board type, size and flags */
    u_char	er_Product;	/* Product number, assigned by manufacturer */
    u_char	er_Flags;	/* Flags */
    u_char	er_Reserved03;	/* Must be zero ($ff inverted) */
    u_short	er_Manufacturer; /* Unique ID,ASSIGNED BY COMMODORE-AMIGA! */
    u_long	er_SerialNumber; /* Available for use by manufacturer */
    u_short	er_InitDiagVec; /* Offset to optional "DiagArea" structure */
    u_char	er_Reserved0c;
    u_char	er_Reserved0d;
    u_char	er_Reserved0e;
    u_char	er_Reserved0f;
};

struct ConfigDev {
    struct Node 	cd_Node;
    u_char		cd_Flags;	/* (read/write) */
    u_char		cd_Pad; 	/* reserved */
    struct ExpansionRom cd_Rom; 	/* copy of board's expansion ROM */
    void		*cd_BoardAddr;	/* where in memory the board was placed */
    u_long		cd_BoardSize;	/* size of board in bytes */
    u_short		cd_SlotAddr;	/* which slot number (PRIVATE) */
    u_short		cd_SlotSize;	/* number of slots (PRIVATE) */
    void		*cd_Driver;	/* pointer to node of driver */
    struct ConfigDev	*cd_NextCD;	/* linked list of drivers to config */
    u_long		cd_Unused[4];	/* for whatever the driver wants */
};

/*
 * Different models of Amiga
 */
#define AMI_UNKNOWN  (0)
#define AMI_500      (1)
#define AMI_2000     (2)
#define AMI_3000     (3)
#define AMI_4000     (4)
#define AMI_1200     (5)

/*
 * chipsets
 */
#define CS_STONEAGE (0)
#define CS_OCS	    (1)
#define CS_ECS	    (2)
#define CS_AGA	    (3)

/*
 * number of CIA Eclock ticks for a single scheduling tick (jiffy)
 */

#define NTSC_JIFFY_TICKS   (7159)
#define PAL_JIFFY_TICKS    (7094)

#endif /* linux/amigatypes.h */
