#ifndef _LINUX_LP_H
#define _LINUX_LP_H

/*
 * usr/include/linux/lp.h  modified for Amiga by Michael Rausch
 *
 * linux i386 version  c.1991-1992 James Wiegand
 * many modifications copyright (C) 1992 Michael K. Johnson
 * Interrupt support added 1993 Nigel Gamble
 */

/*
 * Per POSIX guidelines, this module reserves the LP and lp prefixes
 * These are the lp_table[minor].flags flags...
 */
#define LP_EXIST 0x0001
#define LP_BUSY	 0x0004
#define LP_ABORT 0x0040

/* timeout for each character.  This is relative to bus cycles -- it
 * is the count in a busy loop.  THIS IS THE VALUE TO CHANGE if you
 * have extremely slow printing, or if the machine seems to slow down
 * a lot when you print.  If you have slow printing, increase this
 * number and recompile, and if your system gets bogged down, decrease
 * this number.  This can be changed with the tunelp(8) command as well.
 */

#define LP_INIT_CHAR 400000

/* This is the amount of time that the driver waits for the printer to
 * catch up when the printer's buffer appears to be filled.  If you
 * want to tune this and have a fast printer (i.e. HPIIIP), decrease
 * this number, and if you have a slow printer, increase this number.
 * This is in hundredths of a second, the default 2 being .05 second.
 * Or use the tunelp(8) command, which is especially nice if you want
 * change back and forth between character and graphics printing, which
 * are wildly different...
 */

#define LP_INIT_TIME 2

/* IOCTL numbers */
#define LPCHAR   0x0001  /* corresponds to LP_INIT_CHAR */
#define LPTIME   0x0002  /* corresponds to LP_INIT_TIME */
#define LPABORT  0x0004  /* call with TRUE arg to abort on error,
			    FALSE to retry.  Default is retry.  */
#define LPSETIRQ 0x0005  /* call with new IRQ number,
			    or 0 for polling (no IRQ) */
#define LPGETIRQ 0x0006  /* get the current IRQ number */
#define LPWAIT   0x0008  /* corresponds to LP_INIT_WAIT */

/* timeout for printk'ing a timeout, in jiffies (100ths of a second).
   This is also used for re-checking error conditions if LP_ABORT is
   not set.  This is the default behavior. */

#define LP_TIMEOUT_INTERRUPT	(60 * HZ)
#define LP_TIMEOUT_POLLED	(10 * HZ)

#define LP_F	lp_table.flags		/* flags for busy, etc. */
#define LP_CHAR	lp_table.chars		/* busy timeout */
#define LP_TIME	lp_table.time		/* wait time */
#define LP_WAIT	lp_table.wait		/* strobe wait */
#define LP_IRQ	lp_table.irq		/* interrupt # */
					/* 0 means polled */

#define LP_BUFFER_SIZE 256

struct lp_struct {
	unsigned int irq;
	int flags;
	unsigned int chars;
	unsigned int time;
	struct wait_queue *lp_wait_q;
	char *lp_buffer;
};

struct lp_struct lp_table = {
	0, 0, LP_INIT_CHAR, LP_INIT_TIME, NULL, NULL,
}; 
#define LP_NO 1


#define LP_OUT(c)	ciaa.prb = c

/*  busy is low-active, and so are the cia bits */
#define LP_IS_BUSY	(ciab.pra & 1)
#define LP_HAS_POUT	(ciab.pra & 2)
#define LP_IS_ONLINE	(ciab.pra & 4)

/*
 * function prototypes
 */

extern long lp_init(long);

#endif
