/*
** bootinfo.h -- Defintion of the Linux/68K boot information structure
**
** Copyright 1992 by Greg Harp
**
** This file is subject to the terms and conditions of the GNU General Public
** License.  See the file README.legal in the main directory of this archive
** for more details.
**
** Created 09/29/92 by Greg Harp
*/

#ifndef BOOTINFO_H
#define BOOTINFO_H

#include <linux/amigatypes.h>

/*
 * Amiga specific part of bootinfo structure.
 */

#define NUM_AUTO    16

#define NTSC_ECLOCK 715909
#define PAL_ECLOCK  709379

struct bi_Amiga {
    int
        model;				/* Amiga Model (3000?) */
    int
	num_autocon;			/* # of autoconfig devices found */

    struct ConfigDev
	autocon[NUM_AUTO];		/* up to 16 autoconfig devices */

    u_long
	chip_size;			/* size of chip memory (bytes) */

    u_char
	vblank; 			/* VBLANK frequency */

    u_char
	psfreq; 			/* power supply frequency */

    u_long
	eclock; 			/* EClock frequency */
};


/*
 * CPU and FPU types
 */
#define CPU_68020    (1)
#define CPU_68030    (2)
#define CPU_68040    (4)
#define CPU_MASK     (31)
#define FPU_68881    (32)
#define FPU_68882    (64)

struct mem_info {
    u_long addr;		/* physical address of memory chunk */
    u_long size;		/* length of memory chunk (in bytes) */
};

#define NUM_MEMINFO  4

#define MACH_AMIGA   1
#define MACH_ATARI   2
#define MACH_MAC     3

#define CL_SIZE      (80)

struct bootinfo {
    u_long
	machtype;			/* machine type */

    u_long
	cputype;			/* system CPU & FPU */

    struct mem_info
	memory[NUM_MEMINFO];		/* memory description */

    int
	num_memory;			/* # of memory blocks found */

    u_long
	ramdisk_size;			/* ramdisk size in 1024 byte blocks */

    u_long
	ramdisk_addr;			/* address of the ram disk in mem */

    char
        command_line[CL_SIZE];		/* kernel command line parameters */

    union {
	struct bi_Amiga bi_ami; 	/* Amiga specific information */
    } bi_un;
};
#define bi_amiga bi_un.bi_ami
#define bi_atari bi_un.bi_at
#define bi_mac	 bi_un.bi_m

extern struct bootinfo
    boot_info;

#endif /* BOOTINFO_H */
